ffi.cdef[[
    typedef long LONG;
    typedef void *PVOID;
    typedef unsigned long long ULONG_PTR;
    typedef unsigned long DWORD;
    typedef int BOOL;
    typedef struct _RTL_SRWLOCK {
        PVOID Ptr;
    } RTL_SRWLOCK, *PRTL_SRWLOCK;
    typedef RTL_SRWLOCK SRWLOCK, *PSRWLOCK;

    void InitializeSRWLock(PSRWLOCK SRWLock);
    void AcquireSRWLockExclusive(PSRWLOCK SRWLock);
    void ReleaseSRWLockExclusive(PSRWLOCK SRWLock);
    void AcquireSRWLockShared(PSRWLOCK SRWLock);
    void ReleaseSRWLockShared(PSRWLOCK SRWLock);
    BOOL TryAcquireSRWLockExclusive(PSRWLOCK SRWLock);
    BOOL TryAcquireSRWLockShared(PSRWLOCK SRWLock);
    typedef struct _CONDITION_VARIABLE {
        PVOID Ptr;
    } CONDITION_VARIABLE, *PCONDITION_VARIABLE;


    typedef struct _rw_lock {
        SRWLOCK lock;
        bool    exclusive;
    } rw_lock_t;
]]

local kernel32 = ffi.load"kernel32";
local memory = import"memory"

local function rwlock_new()
    local lock = ffi.cast("rw_lock_t*", memory.alloc(ffi.sizeof("rw_lock_t")))
    kernel32.InitializeSRWLock(lock.lock)
    lock.exclusive = false
    return lock
end

local function rwlock_init(lock)
    kernel32.InitializeSRWLock(lock.lock)
    lock.exclusive = false
end

local function rwlock_delete(lock)
    memory.free(lock.lock)
    memory.free(lock)
end

local function rwlock_tryrlock(lock)
    return kernel32.TryAcquireSRWLockShared(lock.lock) ~= 0
end

local function rwlock_rlock(lock)
    kernel32.AcquireSRWLockShared(lock.lock)
end

local function rwlock_trywlock(lock)
    local ret = kernel32.TryAcquireSRWLockExclusive(lock.lock) ~= 0
    if ret then
        lock.exclusive = true
    end
    return ret
end

local function rwlock_wlock(lock)
    kernel32.AcquireSRWLockExclusive(lock.lock)
    lock.exclusive = true
end

local function rwlock_unlock(lock)
    if lock.exclusive then
        lock.exclusive = false
        kernel32.ReleaseSRWLockExclusive(lock.lock)
    else
        kernel32.ReleaseSRWLockShared(lock.lock)
    end
end

return {
    new = rwlock_new,
    init = rwlock_init,
    rlock = rwlock_rlock,
    wlock = rwlock_wlock,
    unlock = rwlock_unlock,
    delete = rwlock_delete,
    tryrlock = rwlock_tryrlock,
    trywlock = rwlock_trywlock,
}